function [ypr_Bayes mse_Bayes BETA POSTERIOR Test] = LBK_2(grid, X0, Y, mixture_model )
% Implementation of Local bayesian kriging in 1 dimensions
% Input
% grid              : set of positions where prediction is computed (grid.x1, grid.x2)
% X0                : design
% Y                 : observations
% misture_model     : structure defining the mixture model used
%
% Output
% ypr               : predictions over the grid
% Bayes             : 
% mse_Bayes         :
% BETA              :
% POSTERIOR         : 
% Test              :    
% 
%% THERE IS A CHANGE IN COORDINATES BETWEEN INPUT POINTS AND OUTPUT POINTS

% prediction sites
x1 = grid.x;
x2 = grid.y;

n1 = length(x1);
n2 = length(x2);

n0 = length(Y);

% model variables
THETA       = mixture_model.THETA;
LT          = size(THETA, 2);
nu0         = mixture_model.nu0; % parameters for inverse chi-2 prior distribution of sig2
s20         = mixture_model.s20;
p           = mixture_model.p;
np          = mixture_model.np;

F0          = mixture_model.F0;
fx          = mixture_model.fx;
Prior       = mixture_model.Prior;

model0.p        = mixture_model.p;
model0.cov_type = mixture_model.cov_type_0;


model1.p        = mixture_model.p;
model1.cov_type = mixture_model.cov_type_1;

Yx        = zeros(1,LT);
Posterior = zeros(1,LT);
Mse       = zeros(1,LT);
Beta      = zeros(1,LT);
ypr_Bayes = zeros(n1,n2);
mse_Bayes = zeros(n1,n2);

h= waitbar(0,'Computing...');
for i=1:n1
    waitbar(i/n1,h);
    for j=1:n2
        xpr=[x1(i);x2(j)];
        for k=1:LT  % go over model components
            model0.theta = THETA(1,k);
            model1.theta = THETA(2,k);
            % Prediction
            R0      = calcR(X0, model0);
            rx0     = calcrx(xpr, X0, model0);
            rx1     = calcrx(xpr, X0, model1);
            Dm1     = diag(sqrt(rx1));
            F       = Dm1*F0;
            
            beta_hat= ((F'/R0)*F)\((F'/R0)*Dm1*Y);
            yx      = fx*beta_hat + rx0'*(R0\(Dm1*(Y-F0*beta_hat)));
            Yx(k)   = yx;

            n_equiv = n0;
            sig2_hat    = (1/(n_equiv-np))*(Y-F0*beta_hat)'*Dm1*(R0\(Dm1*(Y-F0*beta_hat)));
            % Expected Likelihood with a uniform prior for beta and
            % inv-chi-2 for sig2
            nu0p    = nu0+n_equiv-np;                 
            s20p    = (nu0*s20+(n_equiv-np)*sig2_hat)/(nu0+n_equiv-np);
            ELn     = (det(R0))^(-1/2)*prod(sqrt(rx1))*(det((F'/R0)*F))^(-1/2)*(2*pi)^(-(n_equiv-np)/2)*...
                        gamma(nu0p/2)/gamma(nu0/2)*((s20*nu0/2)^(nu0/2))/((s20p*nu0p/2)^(nu0p/2));
                    
            Posterior(k) = Prior(k)*ELn; 
            
            n_equiv = sum(rx1); % accounts for small number of effective 
                                % observations due to non-stationarity :
                                % only used to compute mse
            sig2_hat= (n0-np)*sig2_hat/(n_equiv-np);
            nu0p    = nu0+n_equiv-np; 
            s20p    = (nu0*s20+(n_equiv-np)*sig2_hat)/(nu0+n_equiv-np);

            
            Mse(k)  = s20p*(1-rx0'*(R0\rx0) + (1-rx0'*(R0\F))^2/((F'/R0)*F) );
            Beta(k) = beta_hat;
        end % k cycle

        Posterior           = Posterior/(sum(Posterior));
        POSTERIOR(j,i,:)    = Posterior;
        ypr_Bayes(i,j)      = Yx*Posterior';
        mse_Bayes(i,j)      = max(0, Mse*Posterior'+((Yx-Yx*Posterior').^2)*Posterior');
    
        Test(i,j)=((Yx-Yx*Posterior').^2)*Posterior';

        BETA(i,j)=Beta*Posterior';
    end % j cycle
end % i cycle
delete(h)

end

%%=========================================================================
function R=calcR(X,model)
%%=========================================================================
% function R=calcR(X,model) 
% R : correlation matrix of the field at the design points (Xi, i = 1:n)
% model.theta & model.p : parameters for correlations R et rx

theta = model.theta;
p     = model.p;

[dimx,n]=size(X);
R=eye(n)/2;
for i=1:n-1
	for j=i+1:n
        if model.cov_type==1 % exponential family + spherical
           S=sqrt(sum(abs(X(:,i)-X(:,j)).^2));
           R(i,j)=exp(-theta*S^p);
        elseif model.cov_type==2 % exponential family, product type (p in [1,2])
           R(i,j)=exp(-theta'*(abs(X(:,i)-X(:,j)).^p))
        elseif model.cov_type==3 %spherical bounded support
           dd=sqrt(sum((X(:,i)-X(:,j)).^2)); 
           R(i,j)=spherical_cov( dd );
        elseif model.cov_type==4  % Stein 1
           pr=prod( ones(dimx,1) + theta.*abs(X(:,i)-X(:,j)));
           R(i,j)=exp(-theta'*(abs(X(:,i)-X(:,j)).^p))*pr;
        elseif model.cov_type==5  % Matern 3/2 + spherical (p=1)
           % = 1/2 for theta*r=1.67835 
           % = 1/5 for theta*r=2.994308347
           S=sqrt(sum(abs(X(:,i)-X(:,j)).^2)); 
           R(i,j)=(1+theta*S)*exp(-theta*S);
        elseif model.cov_type==6  % Matern 5/2 + spherical (p=1)
           % = 1/2 for theta*r=2.33026
           % = 1/5 for theta*r=3.914105418
           S=sqrt(sum(abs(X(:,i)-X(:,j)).^2)); 
           R(i,j)=(1+theta*S+theta^2*S.^2/3)*exp(-theta*S);
        elseif cov_type==7 % Non stationary [Rasmussen, 2006, p 93]
           x0=3*ones(dimx,1);
           Lxi=1-0.75*exp(-8*(abs(X(:,i)-x0).^2)); % dimx times 1
           Lxj=1-0.75*exp(-8*(abs(X(:,j)-x0).^2));
           cc=prod(sqrt(2*Lxi.*Lxj./(Lxi.^2+Lxj.^2)));
           S=sum(abs(X(:,i)-X(:,j)).^2./(Lxi.^2+Lxj.^2));
           R(i,j)=exp(-S)*cc;
        end
	end
end
R=R+triu(R)';

end
%%=========================================================================
function rx = calcrx(x, X, model)
%%=========================================================================
% function rx = calcrx(X,theta,p,x,cov_type) 
% computes the correlations between predicted (x) and design (X) points
% considering model (model)
%
% 
% rx    : vector of correlations between predicted and design points
% model.theta & model.p : correlation function parameters 
% model.cov_type : selects the covariance kernel 
% cov_type==1 : exponential family + spherical
% cov_type==2 : exponential family, product type (p in [1,2])
% cov_type==3 : spherical bounded support
% cov_type==4 : Stein 1
% cov_type==5 : Matern 3/2 + spherical (p=1)
% cov_type==6 : Matern 5/2 + spherical (p=1)
% cov_type==7 : Non stationary [Rasmussen, 2006, p 93]

cov_type = model.cov_type;
theta    = model.theta;
p        = model.p;

[dimx,n]=size(X);

if cov_type==1 % exponential family + spherical    
    S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
    rx = exp(-theta*S.^p)';
    
elseif cov_type==2 % exponential family, product type (p in [1,2])
    rx = exp(-theta'*(abs(X-x*ones(1,n))).^p);
    
elseif cov_type==3 %spherical bounded support
    dd = sqrt(sum((X-x*ones(1,n)).^2))'; 
    rx = spherical_cov( dd );
     
elseif cov_type==4  % Stein 1
     if dimx>1
        pr = prod( ones(dimx,n)+(theta*ones(1,n)).*abs(X-x*ones(1,n)) );
     else
        pr = ones(dimx,n)+(theta*ones(1,n)).*abs(X-x*ones(1,n));
     end
     rx = exp(-theta'*(abs(X-x*ones(1,n))).^p);
     rx = (rx.*pr)';
     
elseif cov_type==5  % Matern 3/2 + spherical (p=1)
     S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
     rx = ((1+theta*S).*exp(-theta*S))';
     
elseif cov_type==6  % Matern 5/2 + spherical (p=1)
     S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
     rx = ((1+theta*S+theta^2*S.^2/3).*exp(-theta*S))';
     
elseif cov_type==7 % Non stationary [Rasmussen, 2006, p 93]
     x0 = 3*ones(dimx,1);
     LX = 1-0.75*exp(-8*(abs(X-x0*ones(1,n)).^2)); % dimx times n
     Lx = 1-0.75*exp(-8*(abs(x-x0).^2));           % dimx times 1
     if dimx>1 
         cc = prod(sqrt(2*LX.*(Lx*ones(1,n))./(LX.^2+(Lx*ones(1,n)).^2)));
     else
         cc = sqrt(2*LX.*(Lx*ones(1,n))./(LX.^2+(Lx*ones(1,n)).^2));
     end
    S  = ones(1,dimx)*((abs(X-x*ones(1,n))).^2./(LX.^2+(Lx*ones(1,n)).^2));
    rx = (exp(-S).*cc)';
end

end

%%=========================================================================

