% Local Bayesian Krigin implementation in dimension 2
% L.Pronzato (I3S/CNRS, France), 2016
% Code corresponds to article "Local Bayesian Kriging"
% Technometrics, 2016.
%
% This script ilustrates the use of Local Bayesian Kriging in dimension
% two.

function RUN_LBK_2()
clear all
close all

% verbose flag
TRACE = 1;

% global cov_type_g % specifies the type of covariance structure
% global cov_type_g p_g X0_g Z0_g F_g
% global model_g;

%__________ SIMULATION__________________________________
% The estimator is tested in a non-stationary process that is the
% point-wise convex combination of two stationary Gaussian processes
%
% Seed points for simulation (of both processes) are a Latin hypercube
% design
d   = 2;
n0  = 10;
X0  = latineHS(d,n0);                
%- - - - - - - - - - - - - - 
% define simulated models
model1.cov_type   = 6;
model1.theta      = 1;
model1.mean       = 0.;
model1.p          = 1;
model1.s2         = 1.;
model1.np         = 1;
model1.fx         = 1.;

model2.cov_type   = 5;
model2.theta      = 7.;
model2.mean       = 10.;
model2.p          = 1;
model2.s2         = 4.;
model2.np         = 1;

% - - - - - - - - - - - - - - 
% mixing function parameters
a    = 30;    % defines the slope of the weighting function
              % large values of a will induce a "strict" boundary
tau1 = 1./2;  % defines the "rectangle" (x>tau1, y> tau2) 
              % where process 2 will have higher weight
tau2 = 1./2;
%- - - - - - - - - - - - - - 
% simulate at X0 the mixed model
% 1st process
[Z0_1 R0_1] = simulateGP(X0, model1);
% 2nd process
[Z0_2 R0_2] = simulateGP(X0, model2);
% True process -> introduction of non-stationarity by weighting
[Z0 m0 w0 C0 IC0] = mix(Z0_1, R0_1, Z0_2, R0_2, model1, model2, X0, a, tau1, tau2);
%
% - - - - - - - - - - - - - - 
% define complete simulation grid
n1  = 41;  n2 = 41; 
x1  = linspace(0, 1, n1);
x2  = linspace(0, 1, n2);
Grid.x = x1;
Grid.y = x2;
Grid.N = n1*n2;
% (matrix of) mixing coefficients for the non-stationary mixing at the
% point of Grid
w    = DefineGridWeights(Grid, a, tau1, tau2);
%
% - - - - - - - - - - - - - - 
% Predict over the simulated grid using model1 w/ ML estimates of its
% parameters (beta and sig2)
Y    = Z0;
F    = ones(n0,1)*model1.fx;                                    % model = cte + gauss processus
R    = calcR(X0, model1);                                       % correlation matrix
beta_hat = ((F'/R)*F)\((F'/R)*Y);                               % Maxium Likelihood estimator of beta
sig2_hat = ((Y-F*beta_hat)'/R)*(Y-F*beta_hat)/(n0-model1.p);    % RML (restricted ML) estimator of sigma2

v    = R\(Y-F*beta_hat);                                        % used for predictor computation ("innovations")
M1   = inv(F'/R*F);
M2   = R\(eye(n0)-F/M1*F'/R);
M3   = R\F*M1;                                              % to compute MSE

% assign space
ypr       = zeros(n2,n1);
MSE_usual = zeros(n2,n1);
Cond_mean  = zeros(n2,n1);
Cond_var   = zeros(n2,n1);
%
for i=1:n1
    for j=1:n2
        xpr     = [x1(i);x2(j)];                % position of the grid point
        rx      = calcrx(xpr, X0, model1);
        fx      = model1.fx;
        yx      = beta_hat + rx'*v;
        
        ypr(j,i) = yx;                          % simulated value at grid point (i,j)
        mse      = sig2_hat*(1-rx'*M2*rx+fx'*M1*fx-2*rx'*M3*fx);
        
        % True conditional mean and variance for the simulation
        k1      = calcrx(xpr, X0, model1);
        k2      = calcrx(xpr, X0, model2);
        kk      = (1-w(j,i))*model1.s2*((-w0+1)'.*k1) + w(j,i)*model2.s2*(w0'.*k2);
        Cond_mean(j,i) = model2.mean*w(j,i) + kk'*IC0*(Z0-m0);
        c0             = (1-w(j,i))^2*model1.s2 + w(j,i)^2*model2.s2;
        Cond_var(j,i)  = c0 - kk'*IC0*kk;
    end
end 
    
% __________ ESTIMATION__________________________________

% chose nonstationary covariance parameter
RULE1 = 0; % use Rule 2
if (RULE1)
    alpha = 10/4;
    aa    = min(1,(alpha*d/n0)^(1/d));
else
    % Rule 2
    phimM       = Minimax_d( X0 );                     % minimax distance criterion
    kNNmax      = max(all_k_NN_distance_d(X0',n0-1));  % row vector of max of kNN distances for k=1:n0
    n_neighbors = min(n0,ceil(10*d/4));
    aa          = kNNmax(n_neighbors-1)+phimM;
end
tt  =   2.99431/aa; % for Matrn 3/2

%- - - - - - - - - - - - - -
% define mixture model for LBK
mixture_model.THETA=[1 5 10 20
                     tt tt tt tt]; %tt=0;   <=> stationary
mixture_model.cov_type_0 = 5; % for K0
mixture_model.cov_type_1 = 5; % for K1
mixture_model.p          = 1; 

mixture_model.nu0 = 2; % parameters for inverse chi-2 prior distribution of sig2
mixture_model.s20 = 1; 

mixture_model.Prior = ones(1,size(mixture_model.THETA,2))/size(mixture_model.THETA,2);    % Prior weigths = uniform
			 
mixture_model.F0  = ones(n0,1); 
mixture_model.fx  = 1;
mixture_model.np  = 1;  % dim(beta)
%- - - - - - - - - - - - - -
% asign space
BAYES = zeros(Grid.N,7);
%% - - - - - - - - - - - - - -
% LBK estimate
[ypr_Bayes mse_Bayes BETA POSTERIOR Test ] = LBK_2(Grid, X0, Y, mixture_model);
    
%
%%__________ GRAPHICAL OUTPUT__________________________________

if TRACE==1
    h = figure(1);
    set(h, 'Position', [60 268 512 384]);
    subplot(221), surf(Grid.x,Grid.y,Cond_mean), shading('flat'), colorbar EastOutside
    shading interp, hold on, plot3(X0(1,:),X0(2,:),Z0,'kp','MarkerSize',5), hold off, title('True conditional mean')
    view(2)
    subplot(222), surf(Grid.x,Grid.y,ypr), shading('flat'), colorbar EastOutside
    shading interp, hold on, plot3(X0(1,:),X0(2,:),Z0,'kp','MarkerSize',5), hold off, title('Observations and classical prediction')
    view(2)
    subplot(223), surf(Grid.x,Grid.y,ypr_Bayes), shading('flat'), colorbar EastOutside
    shading interp, hold on, plot3(X0(1,:),X0(2,:),Z0,'kp','MarkerSize',5), hold off, title('Observations and Bayesian prediction')
    view(2)
    subplot(224), surf(Grid.x, Grid.y, sqrt(Test)), shading('flat'), colorbar EastOutside, shading interp, title('Bayes variability')
    view(2)
    
end

end

%%======================= AUXILIARY FUNCTIONS =============================
function [Z R] = simulateGP(X, model)
%==========================================================================
% simulates the GP over points in X
R = calcR(X, model);	% correlation matrix
Z  = sqrtm(R)*randn(size(X,2),1);

% scale and add mean
Z  = sqrt(model.s2)*Z + model.mean; 


end

%==========================================================================
function w = DefineGridWeights(grid_xy, a, tau1, tau2)
%==========================================================================
fx1 = (exp(a*(grid_xy.x-tau1))./(exp(a*(grid_xy.x-tau1))+exp(-a*(grid_xy.x-tau1))))';
fx2 = exp(a*(grid_xy.y-tau2))./(exp(a*(grid_xy.y-tau2))+exp(-a*(grid_xy.y-tau2)));

% tensor product
w   = fx1*fx2;

end

%==========================================================================
function [Z m w C IC] = mix(Z1, R1, Z2, R2, model1, model2, grid, a, tau1, tau2)
%==========================================================================
% computes a pointwise weighted linear combiation of the two fields 
% Zi (i=1,2) follows modeli, and has covariance matrix Ri 
% on output:
% Z : weighted version
% m : mean
% C : covariance matrix of Z
% IC : the inverse of C


% compute weights
fx1 = exp(a*(grid(1,:)-tau1))./(exp(a*(grid(1,:)-tau1))+exp(-a*(grid(1,:)-tau1)));
fx2 = exp(a*(grid(2,:)-tau2))./(exp(a*(grid(2,:)-tau2))+exp(-a*(grid(2,:)-tau2)));
w   = fx1.*fx2;

Z   = w'.*Z2+(-w+1)'.*Z1; % process
m   = model2.mean*w' + model1.mean*(-w-1)';
C  = model2.s2*(w'*w).*R2 + model1.s2*((-w+1)'*(-w+1)).*R1;
IC = inv(C);


end


%==========================================================================
function k_NNd=all_k_NN_distance_d(X,k)
%==========================================================================
% function k_NNd=all_k_NN_distance_d(X,k)
% for X a N*d matrix, returns the N*k matrix of all jth nearest
% neighbor distances between lines of X

[N,d]=size(X);
if d==1
   k_NNd=all_k_NN_distance(X,k);
   return
end   

% d>1
M=zeros(N,N);
for i=1:d
   Xi=X(:,i);
   Mi=Xi*ones(1,N)-ones(N,1)*Xi';
   Mi=Mi.^2;
   M=M+Mi;
end
MS=sort(M);
k_NNd=sqrt(MS(2:k+1,:))';

end

%==========================================================================
function [ dmm ] = Minimax_d( XX )
%==========================================================================
% function [ dmm ] = Minimax_d( XX )
% computes minimax distance criterion for points in XX (d-dimensional)
% included in the hypercube [0,1]^d
% XX : matrix whose columns are the points x(i), i = 1,..., n

epsi=1e-6;

[d,N0]=size(XX);

%%%%%%%%%%%%%%%% Test if dimension = 1 %%%%%%%%%%%%%
if d==1
    dmm = Minimax_d1( XX, 0,1 );
    return
end    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% compute and include symmetric samples % (d-1)-faces 

for i=1:N0
   Xtst=XX(:,i); % ith point in XX
   for j=1:d
      for k=0:2:2
	      X=Xtst;
   	   X(j)=k-X(j); % symmetry wrt one of the limiting hyperplanes of the cube
         XX=[XX X];
      end
   end
end

% Matlab Delaunay
XX=(unique(XX','rows'))';
if d==2
    TRI=delaunay(XX(1,:),XX(2,:)); 
else
    TRI=delaunayn(XX');
end    

TRI=TRI'; [p,M]=size(TRI);

Centres=zeros(d,M);

r2max=0;
for i=1:M
   [c,r2]=Spheres(XX(:,TRI(:,i)),25);
   Centres(:,i)=c;
   if c>-epsi & c<1+epsi & r2>r2max
		r2max=r2;     
   end
end

dmm=sqrt(r2max);
end

%==========================================================================
function R=calcR(X,model)
%==========================================================================
% function R=calcR(X,model) 
% R : correlation matrix of the field at the design points (Xi, i = 1:n)
% model: strcuture that defines the correlation parameters

theta = model.theta;
p     = model.p;

[dimx,n]=size(X);
R=eye(n)/2;
for i=1:n-1
	for j=i+1:n
        if model.cov_type==1 % exponential family + spherical
           S=sqrt(sum(abs(X(:,i)-X(:,j)).^2));
           R(i,j)=exp(-theta*S^p);
        elseif model.cov_type==2 % exponential family, product type (p in [1,2])
           R(i,j)=exp(-theta'*(abs(X(:,i)-X(:,j)).^p))
        elseif model.cov_type==3 %spherical bounded support
           dd=sqrt(sum((X(:,i)-X(:,j)).^2)); 
           R(i,j)=spherical_cov( dd );
        elseif model.cov_type==4  % Stein 1
           pr=prod( ones(dimx,1) + theta.*abs(X(:,i)-X(:,j)));
           R(i,j)=exp(-theta'*(abs(X(:,i)-X(:,j)).^p))*pr;
        elseif model.cov_type==5  % Matern 3/2 + spherical (p=1)
           % = 1/2 for theta*r=1.67835 
           % = 1/5 for theta*r=2.994308347
           S=sqrt(sum(abs(X(:,i)-X(:,j)).^2)); 
           R(i,j)=(1+theta*S)*exp(-theta*S);
        elseif model.cov_type==6  % Matern 5/2 + spherical (p=1)
           % = 1/2 for theta*r=2.33026
           % = 1/5 for theta*r=3.914105418
           S=sqrt(sum(abs(X(:,i)-X(:,j)).^2)); 
           R(i,j)=(1+theta*S+theta^2*S.^2/3)*exp(-theta*S);
        elseif cov_type==7 % Non stationary [Rasmussen, 2006, p 93]
           x0=3*ones(dimx,1);
           Lxi=1-0.75*exp(-8*(abs(X(:,i)-x0).^2)); % dimx times 1
           Lxj=1-0.75*exp(-8*(abs(X(:,j)-x0).^2));
           cc=prod(sqrt(2*Lxi.*Lxj./(Lxi.^2+Lxj.^2)));
           S=sum(abs(X(:,i)-X(:,j)).^2./(Lxi.^2+Lxj.^2));
           R(i,j)=exp(-S)*cc;
        end
	end
end
R=R+triu(R)';
end

%==========================================================================
function rx = calcrx(x, X, model)
%==========================================================================
% function rx = calcrx(X,theta,p,x,cov_type) 
% computes the correlations between predicted (x) and design (X) points
% considering model (model)
%
% 
% rx    : vector of correlations between predicted and design points
% model.theta & model.p : correlation function parameters 
% model.cov_type : selects the covariance kernel 
% cov_type==1 : exponential family + spherical
% cov_type==2 : exponential family, product type (p in [1,2])
% cov_type==3 : spherical bounded support
% cov_type==4 : Stein 1
% cov_type==5 : Matern 3/2 + spherical (p=1)
% cov_type==6 : Matern 5/2 + spherical (p=1)
% cov_type==7 : Non stationary [Rasmussen, 2006, p 93]

cov_type = model.cov_type;
theta    = model.theta;
p        = model.p;

[dimx,n]=size(X);

if cov_type==1 % exponential family + spherical    
    S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
    rx = exp(-theta*S.^p)';
    
elseif cov_type==2 % exponential family, product type (p in [1,2])
    rx = exp(-theta'*(abs(X-x*ones(1,n))).^p);
    
elseif cov_type==3 %spherical bounded support
    dd = sqrt(sum((X-x*ones(1,n)).^2))'; 
    rx = spherical_cov( dd );
     
elseif cov_type==4  % Stein 1
     if dimx>1
        pr = prod( ones(dimx,n)+(theta*ones(1,n)).*abs(X-x*ones(1,n)) );
     else
        pr = ones(dimx,n)+(theta*ones(1,n)).*abs(X-x*ones(1,n));
     end
     rx = exp(-theta'*(abs(X-x*ones(1,n))).^p);
     rx = (rx.*pr)';
     
elseif cov_type==5  % Matern 3/2 + spherical (p=1)
     S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
     rx = ((1+theta*S).*exp(-theta*S))';
     
elseif cov_type==6  % Matern 5/2 + spherical (p=1)
     S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
     rx = ((1+theta*S+theta^2*S.^2/3).*exp(-theta*S))';
     
elseif cov_type==7 % Non stationary [Rasmussen, 2006, p 93]
     x0 = 3*ones(dimx,1);
     LX = 1-0.75*exp(-8*(abs(X-x0*ones(1,n)).^2)); % dimx times n
     Lx = 1-0.75*exp(-8*(abs(x-x0).^2));           % dimx times 1
     if dimx>1 
         cc = prod(sqrt(2*LX.*(Lx*ones(1,n))./(LX.^2+(Lx*ones(1,n)).^2)));
     else
         cc = sqrt(2*LX.*(Lx*ones(1,n))./(LX.^2+(Lx*ones(1,n)).^2));
     end
    S  = ones(1,dimx)*((abs(X-x*ones(1,n))).^2./(LX.^2+(Lx*ones(1,n)).^2));
    rx = (exp(-S).*cc)';
end

end
%==========================================================================
function [c,beta2]=Spheres(U,nprec)
%==========================================================================
% function [c,beta2]=Spheres(U,nprec)
% for U=[u_1...u_(d+1)] formed of d+1 vectors in R^d
% returns the center of mass c=sum_i (alpha_i u_i) of the u_i's
% and the squared distance ||c-u_i||^2=beta2 for all i
% nprec : precision 1e-nprec on a determinant to consider points as aligned
%
% Constructed from center_mass.m to replace Sphere_old.m

[d,dp]=size(U);

un=ones(d+1,1);
AA=[2*U' -un];
s=pinv(AA)*(sum(U.^2))'; c=s(1:d); beta2=c'*c-s(d+1);

if abs(det(AA))<10^(-nprec)
    c=Inf*ones(d,1);
end   

end

%==========================================================================
function X=latineHS(d,N)
%==========================================================================
% function X=latineHS(d,N)
% returns a design matrix d*N in the hypercube [0,1]^d

[Y,I] = sort(rand(N,d));
I     = I'-ones(d,N);
X     = (rand(d,N)+I)/N;

end
